//Ethan Alexander Shulman "Empty Head" June 15, 2021 
#version 150
precision highp float;

//UNIFORMS/ATTRIBUTES
in vec3 in_Position;
out float time;
out vec4 uv;

uniform sampler2D Buffer0;

layout(std140) uniform ShaderviewBlock {
	vec4 mouse;
	float screenX,screenY;
	int frame,framerate;
};



//CONFIG/CONSTANTS

//Rendering config
//#define DEBUG 1//Debug rendering displays normals. Comment out to disable.
#define STEPS 100//Total path marching steps.
#define EPS 0.1//Marching epsilon.
#define STEP_DITHER 1.//Dither marching step size, hides banding via blurring. Comment out to disable dithering.
#define BACKGROUND_DISTANCE 10.//Distance from geometry before background emission takes effect.
//#define COLOR_MAX 1.//Decreases noise caused by chromatic separation at the cost of accuracy. Comment out to disable.

//Animation config
//#define ANIMATION 1//Toggle animation mode.
//#define FSAMPLES 100//Paths sampled per animation frame.


//CONSTANTS
#define PI 3.14159265


//GLOBAL VARIABLES
vec4 baseRand;//random hash from screen uv and frame
float nsampling;//flag = 1 if sampling normal, used for cancelling volume dither




//UTILITY FUNCTIONS

//2d rotation matrix
mat2 r2(float a) {
	float s = sin(a), c = cos(a);
	return mat2(c,s,-s,c);
}

//tangent and bitangent from normal
mat2x3 tangentBitangent(vec3 norm) {
	mat2x3 t;
	if (abs(norm.x) > abs(norm.y)) t[0] = normalize(vec3(-norm.z,0,norm.x));
	else t[0] = normalize(vec3(0,norm.z,-norm.y)); 
	t[1] = cross(norm,t[0]);
	return t;
}

//uvec4 seed, returns uvec4 hash
uvec4 hash(uvec4 seed) {
    uvec4 h = (0x6A7F8FAAu^seed)*0x01000193u;
    h = ((h.wxyz>>3u)^h^seed.yzwx)*0x01000193u;
    h = ((h.zwxy>>8u)^h^seed.wxyz)*0x01000193u;
    return h^(h>>11u);
}
#define I2F (1./float(0xFFFFFFFFu))


//distance dithering for volumetrics
#define START_VOLUME(posSeed) vec2 opacityRand = fract(baseRand.zw+128.*dot(abs(posSeed),baseRand.zxy)*baseRand.xy)
#define VOLUME(dst,skip,alpha,anisotropy) max(dst,-1e3+(skip+1e3)*step(alpha+nsampling*anisotropy,opacityRand.x))

//dither materials for chromatic separation
#define START_CHROMA(posSeed) float chromaFrame = fract(baseRand.x+156.*dot(abs(posSeed),baseRand.zwy))
#define CHROMA(mat,rc,gc,bc) if (chromaFrame < 1.) {mat[0]*=vec3(3,0,0);rc;}\
else if (chromaFrame < 2.) {mat[0]*=vec3(0,3,0);gc;}\
else {mat[0]*=vec3(0,0,3);bc;}




//PATH TRACING FUNCTIONS

//defines scene geometry, returns distance(x) and material id(y)
vec2 geometry(vec3 p);

/*defines scene materials
mat[0] = color
mat[1].x = roughness, mat[1].y = emission*/
mat2x3 material(vec3 p, float matId);

//returns normal at point
vec3 normal(vec3 p) {
	#define NEPS vec3(EPS,0,0)
	nsampling = 1.;
	vec3 n = normalize(vec3(geometry(p+NEPS).x,geometry(p+NEPS.yxz).x,geometry(p+NEPS.yzx).x)-geometry(p).x);
	nsampling = 0.;
	return n!=n?vec3(0,1,0):n;
}

//defines scene background, returns background color at direction and position
vec3 background(vec3 rp, vec3 rd);

//pathtrace from camera perspective
vec4 pathtrace(vec3 cameraPos, vec3 cameraRot, float camNear, float fovTangent) {
	baseRand = hash(uvec4(uv.yzw,frame))*I2F;
	
	vec3 rp = cameraPos,//ray position
		rd = normalize(vec3(uv.xy+(baseRand.xy*2.-1.)/screenY,fovTangent)),//ray direction
		c = vec3(1),//color tint
		l = vec3(0);//absored light emission

	if (cameraRot.z != 0.) rd.xy *= r2(cameraRot.z);//roll
	if (cameraRot.y != 0.) rd.yz *= r2(cameraRot.y);//pitch
	if (cameraRot.x != 0.) rd.xz *= r2(cameraRot.x);//yaw
	rp += rd*camNear;

	//main path tracing loop
	for (int i = 0; i < STEPS; i++) {
		vec2 d = geometry(rp);
		if (d.x <= EPS) {
			vec4 rand = hash(uvec4(abs(rp)*10000.,frame))*I2F;
		
			//back up ray
			rp -= rd*(EPS*2.-d.x);
			#ifdef DEBUG
			return vec4(normal(rp)*.5+.5,1);
			#endif
			
			//material absorption
			mat2x3 m = material(rp,d.y);
			c *= m[0];
			#ifdef COLOR_MAX
			vec3 normc = c;
			float clen = length(normc);
			if (clen > 1.7320508075688772*COLOR_MAX) normc *= (1.7320508075688772*COLOR_MAX)/clen;
			l += normc*m[1].y;
			#else
			l += c*m[1].y;
			#endif

			//random microsurface bump based on roughness
			float ct = pow(baseRand.x,m[1].x), st = sqrt(1.-ct*ct),
				pt = baseRand.y*PI*2.;
			vec3 nrm = normal(rp);
			mat2x3 tb = tangentBitangent(nrm);
			vec3 bump = normalize(nrm*ct + tb[0]*(st*cos(pt)) + tb[1]*(st*sin(pt)));
			
			//reflect ray
			rd = reflect(rd,bump);
			baseRand.xy = fract(baseRand.xy+baseRand.zw);
		} else {
			#ifdef STEP_DITHER
			rp += rd*max(EPS,d.x-baseRand.z*EPS*STEP_DITHER);
			#else
			rp += rd*d.x;
			#endif
		}
	}
	//add background emission
	l += c*background(rp,rd)*clamp(geometry(rp).x/BACKGROUND_DISTANCE,0.,1.);
	
	return vec4(l,1);
}






//SCENE FUNCTIONS

vec2 geometry(vec3 p) {
	
	float d = abs(length(p)-10.)-.02,//base sphere
		mat = 0.;
		
	//subtract eyes
	vec3 eyep = vec3(abs(p.x),p.yz)-vec3(2,3,-11.);
	eyep.z = abs(eyep.z)-2.;
	float eyed = length(eyep);
	d = max(d,-(eyed-1.5));
	
	eyed -= 1.45;
	if (eyed < d) {
		d = eyed;
		mat = 3.;
	}
	
	//pool
	d = min(d,p.y+7.);
	
	//skull/spine
	vec3 sp = p;
	sp.y = mod(abs(p.y),2.)-1.;
	sp.z -= 8.;
	float sd = min(max(eyed,abs(length(p+sin(32.+p.zyx)*.2)-9.95)-.05),length(sp)-1.4);
	if (sd < d) {
		d = sd;
		mat = 1.;
	}
	
	//pink stuff
	vec3 pp = abs(p-vec3(0,0,0))-3.;
	pp.xz *= r2(.777);
	pp = abs(pp)-1.2;
	pp.yz *= r2(2.);
	
	vec3 pa = abs(pp)-vec3(-3), ba = vec3(3)-vec3(-3);
	float pd = min(length(pp)-.5,length(pa-ba*clamp(dot(pa,ba)/dot(ba,ba),0.,1.))-clamp(abs(p.y)*.05-p.z*.05,0.,1.));
	if (pd < d) {
		d = pd;
		mat = 2.;
	}
	
	return vec2(d,mat);
}

mat2x3 material(vec3 p, float matId) {
	if (matId == 1.) {
		return mat2x3(vec3(.6), vec3(.14,0,0));
	}
	if (matId == 2.) {
		return mat2x3(vec3(.68,.21,.31), vec3(.1,0,0));
	}
	if (matId == 3.) {
		return mat2x3(vec3(1.),vec3(0.,10.,0.));
	}
	return mat2x3(vec3(.3,.01,.005), vec3(.004,0,0));
}

vec3 background(vec3 rp, vec3 rd) {
	return vec3(0);//max(0.,dot(rd,normalize(vec3(0,.8,-3.)))-.92)*500.*vec3(1.,.95,.85);
}



//Vertex and pixel shader
void vert() {
	#ifdef ANIMATION
	time = float(frame)/float(framerate*FSAMPLES);
	#endif
	//uv.xy = aspect correct normalized coordinates, uv.zw = screen pixel coordinates
	uv = vec4(in_Position.xy*vec2(screenX/screenY,1),(in_Position.xy*.5+.5)*vec2(screenX,screenY));
	gl_Position = vec4(in_Position,1);
}
void frag() {
	//progressive rendering
	vec4 l = pathtrace(vec3(-9,0,0), vec3(1.6,0,0),
							 0.,//camera near
							 1.);//field of view
	#ifdef ANIMATION
	gl_FragColor = (frame%FSAMPLES==0?vec4(0):texelFetch(Buffer0,ivec2(uv.zw),0))+l*1e-2;
	#else
	#ifdef VIEW_CONTROLS
	gl_FragColor = (frame==0||mouse.w>0.?vec4(0):texelFetch(Buffer0,ivec2(uv.zw),0))+l*1e-2;
	#else
	gl_FragColor = (frame==0?vec4(0):texelFetch(Buffer0,ivec2(uv.zw),0))+l*1e-2;
	#endif
	#endif
}
